/*
 * Charter.c
 *
 * This file contains all the drawing routines for chart.
 *
 * Bruce Sherrod
 *
 */

#include "header.h"
#include "chart.h"
#include <stdio.h>
#include <math.h>   /* for log10 and ceiling, used for ywidth in vert_output */
#include <string.h> /* for strlen, used for xwidth in horiz_output */
#define toint(x) (int) (x<0?x-0.5:x+0.5)
#define abs(x) (int) (x<0?(-1*x):x)
#define MAX(a,b) a>b?a:b
#define MIN(a,b) a<b?a:b
#define Sprintf (void) sprintf
extern double log10(),pow();

/* shading patterns: */
#define NONE          0        /* no shading, or white			   */
#define BLACK         1        /* full shading, or black                   */
#define GREY          2        /* half shading, or grey                    */
#define HORIZ         3        /* shaded with horizontal lines		   */
#define VERT	      4        /* shaded with vertical lines		   */
#define CROSS	      5        /* cross-hatched; HORIZ and VERT		   */
#define DIAG          6        /* diagonal lines with slope 1              */
#define NEG_DIAG      7        /* diagonal lines with slope -1             */
#define CROSS_DIAG    8        /* cross-hatched; DIAG and NEG_DIAG         */
			      
 
void picstart()
/* write the gunk required at the beginning of the picture */
{
    int pensize = 8;  /* bug/possible new feature: this should
			 probably be user-settable. */
 
    /* not sure what all this does, but it seems necessary... */
    Fprintf(outfile,"\\catcode`@=11\n");
    Fprintf(outfile,"\\expandafter\\ifx\\csname graph\\endcsname\\relax \\alloc@4\\box\\chardef\\insc@unt\\graph\\fi\n");
    Fprintf(outfile,"\\catcode`@=12\n");
    Fprintf(outfile,"\\setbox\\graph=\\vtop{%\n");
    Fprintf(outfile,"  \\baselineskip=0pt \\lineskip=0pt \\lineskiplimit=0pt\n");
    Fprintf(outfile,"  \\vbox to0pt{\\hbox{%\n");
    Fprintf(outfile,"    \\special{pn %d}%\n",pensize); 
}
 
 
void picend()
/* write the gunk required at the end of the picture */
{

    /* adjust the picture so that it is centered... */
    Fprintf(outfile,"    \\kern %fin\n",totalwidth);
    Fprintf(outfile,"  }\\vss}%\n");
    Fprintf(outfile,"  \\kern %fin\n",totalheight);
    Fprintf(outfile,"}\n");
    
}
 
 
void drawline(x1,y1,x2,y2)
int x1,y1,x2,y2;
/* draw a line from (tex coords!) (x1,y1) to (x2,y2) */
{
    Fprintf(outfile,"    \\special{pa %d %d}%%\n",x1,y1);
    Fprintf(outfile,"    \\special{pa %d %d}%%\n",x2,y2);
    Fprintf(outfile,"    \\special{fp}%%\n");   
}
		 
 
void drawtext(x,y,text,center)
int x,y;
char *text;
int center;
/* "draw" the text at location x,y (tex coords).
   Since x and y are in milli-inches, they need to be converted to inches.
   if center = 0, then text is centered around (x,y).  
   if center = 1, then text is placed to the left of (x,y);
   if center = 2, then text is placed to the right of (x,y).
*/
{
    Fprintf(outfile,"    \\rlap{\\kern %3.2fin\\lower %3.2fin\\hbox to 0pt{%s %s %s}}%%\n",(double) x/1000,(double) y/1000,
	    center==0||center==1?"\\hss":"",
	    text,
	    center==0||center==2?"\\hss":"");
}
 
 
void drawrec(x1,y1,x2,y2,shade)
int x1,y1,x2,y2,shade;
/* draw a rectangle with (x1,y1) as the upper left corner, (x2,y2) as
   the lower right corner, and shade as the shading pattern. */
{
    /* This function assumes that (x1,y1) is the upper left corner.
       However, in some cases, it won't be.  If it isn't, we'll make it so. */
    if (y1 > y2) { int temp=y1; y1=y2; y2=temp; }  /* swap them */
    if (x1 > x2) { int temp=x1; x1=x2; x2=temp; }  /* swap them */	
       
    /* outline the rectangle */
    Fprintf(outfile,"    \\special{pa %d %d}%%\n",x1,y1);
    Fprintf(outfile,"    \\special{pa %d %d}%%\n",x2,y1);
    Fprintf(outfile,"    \\special{pa %d %d}%%\n",x2,y2);
    Fprintf(outfile,"    \\special{pa %d %d}%%\n",x1,y2);
    Fprintf(outfile,"    \\special{pa %d %d}%%\n",x1,y1);
 
    /* shading */
    if (shade == GREY)
	Fprintf(outfile,"    \\special{sh 0.5}%%\n");
    else if (shade == BLACK)
	Fprintf(outfile,"    \\special{sh 1}%%\n");
    else
	Fprintf(outfile,"    \\special{sh 0}%%\n");
 
    Fprintf(outfile,"    \\special{fp}%%\n");  /* finish the rectangle */

    if ((shade == HORIZ) || (shade == CROSS)) {
    /* shading with horizontal lines */
        int i=0;
        for(i=y1;i<y2;i+=shadedensity[shade]) drawline(x1,i,x2,i);
    }
    
    if ((shade == VERT) || (shade == CROSS)) {
    /* shading with vertical lines */
        int i=0;
        for(i=x1;i<x2;i+=shadedensity[shade]) drawline(i,y1,i,y2);
    }
    
    if ((shade == DIAG) || (shade == CROSS_DIAG)) {
    /* Diagonal shading with positive slope.  Positive means
       from the lower left to the upper right; in tex coordinates,
       a line with "positive" slope with have the form (x,-x+b). */
        int i;
	
	for(i=y1;i<=y2+(x2-x1);i+=shadedensity[shade]) {
	    /* defines lines with the form (x1+a,i-a). */
	    int b1,b2,b3,temp_x1=x1,temp_i=i;
	    b1 = x1+i-y1; /* intersection with the top; form (b1,y1). */
	    b2 = i-x2+x1; /* intersection with the right side; form (x2,b2). */
	    b3 = x1+i-y2; /* intersection with the bottom; form (b3,y2) */
 
	    if (i>=y2) { temp_x1 = b3; temp_i = y2; } 
	    if (b1<=x2) drawline(temp_x1,temp_i,b1,y1); 
	    else drawline(temp_x1,temp_i,x2,b2); 
	}
    }
 
    if ((shade == NEG_DIAG) || (shade == CROSS_DIAG)) {
    /* Diagonal shading with negative slope */
	int i;
 
	for(i=y1;i<=y2+(x2-x1);i+=shadedensity[shade]){
	    /* defines lines of the form (x2+a,i+a) */
	    int b1,b2,b3,temp_x2=x2,temp_i=i;
	    b1 = x2+y1-i; /* intersection with top; form (b1,y1). */
	    b2 = i+x1-x2; /* intersection with left side; form (x1,b2). */
	    b3 = x2+y2-i; /* intersection with bottom; form (b3,y2). */
	    
	    if (i>=y2) { temp_x2 = b3; temp_i = y2; } 
	    if (b1<=x1) drawline(temp_x2,temp_i,x1,b2);
	    else drawline(temp_x2,temp_i,b1,y1); 
	}
    }
    
} /* drawrec */
 
 
/*
   assign uperin and inperu, as well as max, min, and anything else
   that looks like it needs to be set.  Divisor is passed by reference
   for practice (no reason).
*/
void choose_uperin(divisor)
double *divisor;
{
    int i1,i2;
 
    min = max = 0.0;
    for (i1=0; i1<=nsets; i1++)
	for (i2=0; i2<npoints[i1]; i2++) {
	    if (y[i1][i2] < min) min = y[i1][i2];
	    if (y[i1][i2] > max) max = y[i1][i2];
	}
    if (min > 0.0) min = 0.0;
    if (max < 0.0) max = 0.0;
    
    /* This chooses a value for interval, if it is
       not given.  It should choose fairly nice intervals.
       The concept here is that the
           10 ^ *truncated* log10 of the ugly interval
       is a nice number.  In addition, we preserve the most significant
       digit.  The ugly interval is an interval that corresponds
       to 6 ticks.  If the user gives a bogus interval, it will not be
       noticed.	*/
    if (interval==-1) {
	int temp_interval,size;
	temp_interval = (max-min)/6;
	size = toint(log10((double) temp_interval));
	interval = toint( temp_interval/pow(10.0,(double) size) );
	interval = toint( interval*pow(10.0,(double) size) );
#ifndef	notdef
	if (interval == 0) interval = temp_interval;
#endif
    }

    uperin =(max-min) / *divisor;
    inperu = 1.0 / uperin;
    
}
 
 
void vertical_output()
/* draw a vertical bar chart */
{
    int i,cur,curset,mymax;         /* current data, point and set */
    char *tmp="xxxxxx";       /* temp. space to convert int to str */
    int ywidth;		      /* size of the largest number on the Y axis;
				 in milli-inches; used for scaling */
    int bigset=0;	      /* the set with the most points. */

    /* this assures that if all the data sets do not have the same
       number of points, then the largest number of points is used. */
    for(i=0;i<=nsets;i++) if (npoints[i]>bigset) bigset=i;
 
    /* this causes the entire graph to be drawn a bit smaller, insuring
       that space is left for the msg's and axis labels. */
    switch (label) {
        case BOTTOM:
	    totalheight -= nsets==0?100/1000.0:(nsets*100+100)/1000.0;
	    break;
	case NEAR:
	    totalheight -= 300/1000.0;
	    break;
    }
    totalheight -= tsep/1000.0;
    choose_uperin(&totalheight);
    ywidth = toint(ceil(log10( MAX(max,-1*min) ))*70);
    totalwidth-=(ywidth+10)/1000.0;

    /* the icky part. This section sets sep and/or bar_thickness */
    if (bar_thickness == 0.0) {
	if (sep == 0.0)
	    sep = bar_thickness = (totalwidth*PTPERIN-npoints[bigset]*nsets*offset)/((nsets+2)*npoints[bigset]);
	else
	    bar_thickness = (totalwidth*PTPERIN-npoints[bigset]*nsets*offset-npoints[bigset]*sep)/((nsets+1)*npoints[bigset]);
    } else if (sep == 0.0)
	sep = (totalwidth*PTPERIN-npoints[bigset]*nsets*offset-(nsets+1)*bar_thickness*npoints[bigset])/npoints[bigset];
    else /* user set both.  If the sep and bar_thickness are bogus,
	    adjust the totalwidth to account for it. */
       	if (npoints[bigset]*sep+(nsets+1)*bar_thickness*npoints[bigset]+npoints[bigset]*nsets*offset > totalwidth) totalwidth = (npoints[bigset]*sep+(nsets+1)*bar_thickness*npoints[bigset]+npoints[bigset]*nsets*offset)/1000.0;

#ifdef DEBUG
    dump_info(); 
#endif
    
    mymax = max / interval;
    mymax *= interval;
    if (mymax < max)
	mymax += interval;
    if (mymax > max)
	totalheight += (mymax - max) * inperu;

    /* draw the Y axis */
    if (max > 0) { /* draw the positive part of the Y axis */
        for (cur=0; cur<=mymax; cur += interval) {
	    drawline(ywidth, (int) ((mymax-cur)*inperu*PTPERIN),
		     ywidth+30, (int) ((mymax-cur)*inperu*PTPERIN));
		     /* tick mark */
	    Sprintf(tmp,"%d",cur);
	    drawtext(ywidth-10,(int)((mymax-cur)*inperu*PTPERIN),tmp,1);
		    /* label tick mark */
	    if (cur>0) drawline(ywidth,toint((mymax-cur)*inperu*PTPERIN),
			      ywidth,toint((mymax-cur+interval)*inperu*PTPERIN));
	       	    /* line to next tick mark; these lines go the opposite
		       direction of the for loop (from larger to smaller) */
	}
    }
    if (min < 0) { /* draw the negative part of the Y axis */
        for (cur=0; cur>min; cur -= interval) {
	    drawline(ywidth, (int) ((mymax-cur)*inperu*PTPERIN),
	             ywidth+30, (int) ((mymax-cur)*inperu*PTPERIN));
	    /* tick mark */
	    Sprintf(tmp,"%d",cur);
	    drawtext(ywidth-10,(int)((mymax-cur)*inperu*PTPERIN),tmp,1);
	    /* label tick mark */
	    if (cur>min+interval)
		drawline(ywidth,toint((mymax-cur)*inperu*PTPERIN),
			 ywidth,toint((mymax-cur+interval)*inperu*PTPERIN));
	    /* line to next tick mark; these lines go the opposite
	       direction of the for loop (from larger to smaller) */
	}
	/* draw a segment from the last label to the bottom of the graph,
	   and, if it needs it, label it.  A label is applied if it can
	   fit (without overlapping the previous label. */
	cur = mymax-min;
	if ( (abs(toint(min))%interval)*inperu > 0.15) {
	    drawline(ywidth,toint(cur*inperu*PTPERIN),
		     ywidth+30,toint(cur*inperu*PTPERIN)); /* tick mark */
	    Sprintf(tmp,"%d",toint(min));
	    drawtext(ywidth-10,toint(cur*inperu*PTPERIN),tmp,1);
	}
	drawline(ywidth,toint(cur*inperu*PTPERIN),
		 ywidth,toint((cur-interval)*inperu*PTPERIN));
	    /* line to previous mark */
    }
    
    /* draw the X axis */
    drawline(ywidth, toint(mymax*inperu*PTPERIN),
	     (int) (totalwidth*PTPERIN),toint(mymax*inperu*PTPERIN));
    
    /* draw the bars */
    curset =0; cur=0; i=sep+ywidth;
    while(curset<=nsets && cur < npoints[bigset]){
	if (cur<npoints[curset]) {
		drawrec(i,toint((mymax-y[curset][cur])*inperu*PTPERIN),
			toint(i+bar_thickness),toint(mymax*inperu*PTPERIN),
			shade[curset]);
		switch (label) {
		    case BOTTOM:
		        drawtext(toint(i+bar_thickness/2),
				 toint(totalheight*PTPERIN+tsep+curset*110),
				 msg[curset][cur],0);
			break;
		    case NEAR:
			drawtext(toint(i+bar_thickness/2),
				 y[curset][cur]>0?toint((mymax-y[curset][cur])*inperu*PTPERIN-tsep):toint((mymax-y[curset][cur])*inperu*PTPERIN+tsep),
				 msg[curset][cur],0);
			break;
	        }
        }
	i+=bar_thickness;
	if (curset<nsets) { curset++; i+= offset; }
	else { curset=0; i+=sep; cur++; }
    }
 
    /* restore the size of the graph. */
    switch (label) {
        case BOTTOM:
	    totalheight += nsets==0?100/1000.0:(nsets*100+100)/1000.0;
	    break;
	case NEAR:
	    totalheight+=300/1000.0;
	    break;
    }
    totalheight+=tsep/1000.0;
    totalwidth+=(ywidth+10)/1000.0;
} /* vertical_output */
 
 
void horizontal_output()
/* draw a horizontal bar chart */
{
    int i,cur,curset;         /* current data, point and set */
    char *tmp="xxxxxx";       /* temp. space to convert int to str */
    int xwidth=0;      	      /* size of the largest msg on the X axis;
				 in milli-inches; used for scaling */
    int bigset=0;	      /* the set with the most points. */
    
    /* This allows for sets of unequal sizes */
    for(i=0;i<=nsets;i++) if (npoints[i]>bigset) bigset=i;
    
    /* this causes the entire graph to be drawn a bit smaller, insuring
       that space is left for the msg's and axis labels. */
    totalheight-=90/1000.0;
 
    /* this sets xwidth to the width of the largest msg, in pts,
       where each character is considered to be 70 milli-inches wide.
       That is a little wide for letters, but correct for numbers. */
    for (curset=0;curset<=nsets;curset++)
	for(cur=0;cur<npoints[curset];cur++)
	    if( xwidth<strlen(msg[curset][cur]) )
		xwidth = strlen(msg[curset][cur]);
    xwidth *= 70; /* the width of a number.  Note: if the msg's are
		     primarily text, this will be a bit large. */
    xwidth += tsep;

    switch (label) {
        case BOTTOM:    
            totalwidth-=xwidth/1000.0; /* the width of the largest msg */
	    break;
	case NEAR:
            totalwidth-=(2*xwidth)/1000.0;
	     /* the width of the largest msg, on each side */  
	    break;
    }
    choose_uperin(&totalwidth);
    if (label==NEAR) totalwidth += xwidth/1000.0;
    /* restore the xwidth that the chart uses */
	
    /* fix to allow for negative values: note that this doesn't
       work for bar labeling, so there is an exception there. */
    xwidth += min<0?abs(min*inperu*PTPERIN):0;
    
    if (bar_thickness == 0.0) {
	if (sep == 0.0)
	    sep = bar_thickness = (totalheight*PTPERIN-npoints[bigset]*nsets*offset)/((nsets+2)*npoints[bigset]);
	else
	    bar_thickness = (totalheight*PTPERIN-npoints[bigset]*nsets*offset-npoints[bigset]*sep)/((nsets+1)*npoints[bigset]);
    } else if (sep == 0.0)
	sep = (totalheight*PTPERIN-npoints[bigset]*nsets*offset-(nsets+1)*bar_thickness*npoints[bigset])/npoints[bigset];
    else /* both are set.  If the user enters bogus bar_thickness and
	  sep, then readjust the totalheight. */
	if (npoints[bigset]*sep+(nsets+1)*bar_thickness*npoints[bigset]+npoints[bigset]*nsets*offset > totalheight)
	    totalheight = (npoints[bigset]*sep+(nsets+1)*bar_thickness*npoints[bigset]+npoints[bigset]*nsets*offset)/1000.0;
 
#ifdef DEBUG
    dump_info();
#endif
    
    /* draw the X axis */
    if (max > 0) { /* draw the positive part of the X axis */
        for (cur=0; cur<max; cur += interval) {
	    drawline(toint(cur*inperu*PTPERIN+xwidth),70,
		     toint(cur*inperu*PTPERIN+xwidth),100);
	    /* tick mark */
	    Sprintf(tmp,"%d",cur);
	    drawtext(toint(cur*inperu*PTPERIN+xwidth),50,tmp,0);
	    /* label tick mark */
	    if (cur+interval<max)
		drawline(toint(cur*inperu*PTPERIN+xwidth),70,
			 toint((cur+interval)*inperu*PTPERIN)+xwidth,70);
	        /* line to next tick mark */
	}
	/* draw a segment from the last label to the top of the graph,
	   and, if it needs it, label it.  A label is applied if it can
	   fit (without overlapping the previous label. */
	cur -= interval;
	if ( (max-cur)*inperu*PTPERIN > log10((double)interval)*70) {
	drawline(toint(max*inperu*PTPERIN+xwidth),70,
		 toint(max*inperu*PTPERIN+xwidth),100);
		/* tick mark */
	    Sprintf(tmp,"%d",toint(max));
	    drawtext(toint(max*inperu*PTPERIN)+xwidth,50,tmp,0);
	}
	drawline(toint(cur*inperu*PTPERIN+xwidth),70,
		 toint(max*inperu*PTPERIN+xwidth),70);
	    /* line to previous mark */
    }
    if (min < 0) { /* draw the negative part of the X axis */
        for (cur=0; cur>min; cur -= interval) {
	    drawline(toint(cur*inperu*PTPERIN+xwidth),70,
		     toint(cur*inperu*PTPERIN+xwidth),100);
	    /* tick mark */
	    Sprintf(tmp,"%d",cur);
	    drawtext(toint(cur*inperu*PTPERIN+xwidth),50,tmp,0);
	    /* label tick mark */
	    if (cur-interval>min)
		drawline(toint(cur*inperu*PTPERIN+xwidth),70,
			 toint((cur-interval)*inperu*PTPERIN+xwidth),70);
	        /* line to next tick mark */
	}	
	/* draw a segment from the last label to the top of the graph,
	   and, if it needs it, label it.  A label is applied if it can
	   fit (without overlapping the previous label. */
	cur += interval;
	if ( toint((cur-min)*inperu*PTPERIN>log10((double)interval)*50)) {
	drawline(toint(min*inperu*PTPERIN+xwidth),70,
		 toint(min*inperu*PTPERIN+xwidth),100);
		/* tick mark */
	    Sprintf(tmp,"%d",toint(min));
	    drawtext(toint(min*inperu*PTPERIN)+xwidth,50,tmp,0);
	}
	drawline(toint(cur*inperu*PTPERIN+xwidth),70,
		 toint(min*inperu*PTPERIN+xwidth),70);
	    /* line to previous mark */
    }

    /* draw the Y axis */
    drawline(xwidth,70,xwidth,toint(totalheight*PTPERIN)); 
    
    /* draw the bars */
    curset =0; cur=0; i=sep+70;
    while(curset<=nsets && cur < npoints[bigset]){
	if (cur<npoints[curset]) {
		drawrec(xwidth,i,
			toint(y[curset][cur]*inperu*PTPERIN+xwidth),
			toint(i+bar_thickness),
			shade[curset]);
		switch (label) {
		    case BOTTOM:
		        drawtext(toint(xwidth-(min<0?
					    abs(min*inperu*PTPERIN):0)-tsep),
			     toint(i+bar_thickness/2+35),
			     msg[curset][cur],1);
		    /* note the anti-fix of xwidth for negative values. */
			break;
		    case NEAR:
			drawtext(y[curset][cur]<0?
			    toint(y[curset][cur]*inperu*PTPERIN+xwidth-tsep):
			    toint(y[curset][cur]*inperu*PTPERIN+xwidth+tsep),
			    toint(i+bar_thickness/2+35),
			    msg[curset][cur],y[curset][cur]<0?1:2);
			break;
		}
        }
	i+=bar_thickness;
	if (curset<nsets) { curset++; i+= offset; }
	else { curset=0; i+=sep; cur++; }
    }

    xwidth -= min<0?abs(min*inperu*PTPERIN):0; 

    /* restore the size of the graph. */
    totalheight+=90/1000.0;
    totalwidth+=xwidth/1000.0;
    
} /* horizontal_output */
 
 
 
/* This routine produces tex input for a pie chart of ONE data set.
   This routine was copied (almost) directly from the original chart. 
*/
void pie_output(set)
int set;
{
    double percent[MAXPOINTS], total=0.0, angle, tangle, hradius;
    register int i, r, ix, iy;

    /* guarantee that the pie chart is actually within the box */
    totalheight = 2*radius;
    totalwidth = 0;
    
    for (i=0; i<npoints[set]; i++) {
	if (y[set][i] < 0.0) {
	    Fprintf(stderr, "Values cannot be < 0 for pie charts\n");
	    exit(1);
	}
	total += y[set][i];
    }
    for (i=0; i<npoints[set]; i++)
	percent[i] = y[set][i] / total;
    r = radius * 1000.0 + 0.5;
    hradius = radius / 2.0;

#ifdef DEBUG
    dump_info();
#endif
    
    Fprintf(outfile, "\\vtop{\\vbox to 0pt{\\hbox{%%\n");
    Fprintf(outfile, "  \\special{ar %d %d %d %d 0 6.283}%%\n",
	r, r, r, r);
    for (angle=0, i=0; i<npoints[set]; i++) {
	tangle = angle + percent[i] * TWOPI / 2.0;
	ix = radius * sin(angle) * 1000.0 + 0.5;
	iy = radius * cos(angle) * 1000.0 + 0.5;
	Fprintf(outfile,
	    "  \\special{pa %d %d}\\special{pa %d %d}\\special{fp}%%\n",
	    r, r, r+ix, r+iy);
	Fprintf(outfile,
	    "  \\rlap{\\kern %gin\\lower %gin\\hbox to 0pt{\\hss %s\\hss}}%%\n"
		,hradius * sin(tangle) + radius,
		hradius * cos(tangle) + radius,
		msg[set][i]);
	angle += percent[i] * TWOPI;
    }
    Fprintf(outfile, "\\kern %gin}\\vss}\\kern %gin}%%\n",
	radius*2.0, radius*2.0);
}

